structure Main =
struct

type int = Int.t

structure Color = Color ()
structure Document = Document (structure Color = Color)
structure Meaning = Meaning (structure Color = Color
			     structure Document = Document)
structure GenRandom = GenRandom (structure Document = Document
				 structure Meaning = Meaning)
structure Doit = Doit (structure Document = Document
		       structure Meaning = Meaning)
structure Eval = Doit.Eval
structure Parse = Doit.Parse
structure Peephole = Doit.Peephole
structure Synthesize = Synthesize(structure Document = Document
				  structure Meaning = Meaning)
   
fun usage s =
   Process.usage {usage = "[-mode {equiv|eval|optimize|parse|parseNoPrint|parseSize}] timeLimit",
		  msg = s}
fun genrandom1 args =
    case args of
	[s] =>
	let val size =
		case Int.fromString s of
		    NONE => usage "random1 expects an integer"
		  | SOME s => s
	    val _ = Random.srand (Random.seed ())
	    val d = GenRandom.generate1 size
	in  Document.output (d, Out.standard)
	end
      | _ => usage "random1 expects a size argument"

fun genrandom2 args =
    case args of
	[s] =>
	let val size =
		case Int.fromString s of
		    NONE => usage "random2 expects an integer"
		  | SOME s => s
	    val _ = Random.srand (Random.seed ())
	    val m = GenRandom.generate2 size
	    val d = Synthesize.synthesize1 m
	in  Document.output (d, Out.standard)
	end
      | _ => usage "random1 expects a size argument"

fun equiv args =
   case args of
      [f1, f2] =>
	 let
	    fun z f = Eval.eval (Parse.parse f)
	    val m1 = z f1
	    val m2 = z f2
	    fun msg m =
	       Layout.outputl (Meaning.layout m, Out.error)
	 in
	    if Meaning.equals (m1, m2)
	       then ()
	    else
	       (Meaning.explainNonEquals (m1, m2, Out.error)
		; raise Fail "equiv failed")
	 end
    | _ => usage "equiv expects two args"

fun evalNoPrint args =
   case args of
      [f] => (Eval.eval (Parse.parse f); ())
    | _ => usage "evalNoPrint expects one arg"

fun eval args =
   case args of
      [f] =>
	 Layout.outputl (Meaning.layout (Eval.eval (Parse.parse f)),
			 Out.standard)
    | _ => usage "eval expects one arg"

fun optimize args =
   case args of
      [timeLimit] =>
	 (case Int.fromString timeLimit of
	     NONE => usage "timeLimit must be an integer"
	   | SOME t => 
		Doit.doit (Time.seconds (LargeInt.fromInt t),
			   In.standard, Out.standard))
    | _ => usage "wrong number of args"

fun parseNoPrint args =
   case args of
      [f] => (Parse.parse f; ())
    | _ => usage "parseNoPrint expects one arg"

fun parseSize args =
   case args of
      [f] =>
	 let
	    val t = Parse.parse f
	    val _ = print (concat ["Tree size is ",
				   Int.toString (MLton.size t), "\n"])
	 in
	    ()
	 end
    | _ => usage "parseSize expects one arg"

fun parse args =
   case args of
      [f] => Document.output (Parse.parse f, Out.standard)
    | _ => usage "parse expects one arg"

fun peephole args =
   case args of
      [f] =>
	let
	  val d = Parse.parse f
	  val m = Eval.eval d
	  val d' = Peephole.peephole d
	  val m' = Eval.eval d'
	  val _ = Document.output (d', Out.standard)
	in
	  if Meaning.equals(m, m')
	    then ()
	    else
	       (Meaning.explainNonEquals (m, m', Out.error)
		; raise Fail "peephole failed")
	 end
    | _ => usage "peephole expects one arg"

fun strip args =
   case args of
      [f] => (Parse.strip f; ())
    | _ => usage "strip expects one arg"

structure Token = Parse.Token
fun tokenize args =
   case args of
      [f] =>
	 let
	    val nextToken = Token.tokenize f
	    fun loop () =
	       if Token.isEof (nextToken ())
		  then ()
	       else loop ()
	 in
	    loop ()
	 end
    | _ => usage "tokenize expects one arg"

val mode = ref optimize

fun main args =
   let
      val rest =
	 Popt.parse
	 {switches = args,
	  opts = [("mode",
		   Popt.SpaceString
		   (fn s =>
		    mode := (case s of
				"eval" => eval
			      | "evalNoPrint" => evalNoPrint
			      | "equiv" => equiv
			      | "optimize" => optimize
			      | "parse" => parse
			      | "parseNoPrint" => parseNoPrint
			      | "parseSize" => parseSize
			      | "peephole" => peephole
			      | "random1" => genrandom1
			      | "random2" => genrandom2
			      | "strip" => strip
			      | "tokenize" => tokenize
			      | _ => usage "strange mode"))),
		  Popt.trace]}
    in
      case rest 
	of Result.No s => usage (concat ["invalid switch: ", s])
	 | Result.Yes args => !mode args
   end

fun exportNJ (file: File.t): unit =
   SMLofNJ.exportFn
   (file,
    fn (_, args) => ((main args; OS.Process.success)
		     handle e =>
			let open Layout
			in output (seq [str "main: ", Exn.layout e],
				   Out.error)
			   ; Out.newline Out.error
			   ; OS.Process.failure
			end))

val main = Process.makeMain main

end
