type int = Int.t

functor Meaning (S: MEANING_STRUCTS): MEANING =
struct

open S
   
structure Proprec =
   struct
      type t = {b: bool,
		color: Color.t,
		em: bool,
		i: bool,
		s: bool,
		size: int option,
		tt: bool,
		u: int}

      fun layout ({b, color, em, i, s, size, tt, u}: t) =
	 Layout.record [("b", Bool.layout b),
			("color", Color.layout color),
			("em", Bool.layout em),
			("i", Bool.layout i),
			("s", Bool.layout s),
			("size", Option.layout Int.layout size),
			("tt", Bool.layout tt),
			("u", Int.layout u)]

      val equals: t * t -> bool = op =

      val equals =
	 Trace.trace2 ("Proprec.equals", layout, layout, Bool.layout) equals

      val root: t =
	 {u = 0, b = false, em = false, i = false, s = false, tt = false,
	  size = NONE,
	  color = Color.Unknown}

      fun space (ctxt as {b, em , i, s, tt, u, size, color}: t)
	= {b = false, em = false, i = false, s = false, tt = tt, u = u,
	   color = if u = 0 then Color.White else color,
	   size = size}

      val toString = Layout.toString o layout

      fun update (ctxt as {b, em , i, s, tt, u, size, color}: t,
		  tag: Document.tag) : t =
	 let datatype z = datatype Document.tag
	 in  case tag of
	    B => {b = true,
		  em = em, i = i, s = s, tt = tt, u = u,
		  size = size, color = color}
	  | EM => if s
		     then ctxt
		  else {em = not em,
			b = b, i = i, s = s, tt = tt, u = u,
			size = size, color = color}
	  | I => {i = true,
		  em = em, b = b, s = s, tt = tt, u = u,
		  size = size, color = color}
	  | PL => {i = false, s = false, em = false,
		   b = false, tt = false, u = 0,
		   size = size, color = color}
	  | S => {s = true, em = false,
		  i = i, b = b, tt = tt, u = u,
		  size = size, color = color}
	  | TT => {tt = true,
		   b = b, em = em, i = i, s = s, u = u,
		   size = size, color = color}
	  | U => if u < 3
		    then {u = u + 1,
			  b = b, em = em, i = i, s = s, tt = tt,
			  size = size, color = color}
		 else ctxt
	  | Size size => {size = SOME size,
			  b = b, em = em, i = i, s = s, tt = tt,
			  u = u, color = color}
	  | Color color => {color = color,
			    b = b, em = em, i = i, s = s, tt = tt,
			    u = u, size = size}
	 end
   end

structure External =
   struct
      type t = { c: char, p: Proprec.t }

      fun layout ({c, p}: t): Layout.t =
	 Layout.record [("c", Int.layout (Char.toInt c)),
			("p", Proprec.layout p)]
   end

type decchar = External.t
(* Internally a decorated character is represented as a
n integer as follows:
 *   bit     24: b
 *           23: em
 *           22: i
 *           21: s
 *           20: tt
 *   bits 18-19: u
 *   bits 14-17: size
 *   bits 11-13: color
 *   bits  0- 7: c
 *)
type internal = int
     
type t = internal array

fun andb (i1,i2) = Word.toInt (Word.andb (Word.fromInt i1,Word.fromInt i2))
fun orb (i1,i2) = Word.toInt (Word.orb (Word.fromInt i1,Word.fromInt i2))
fun shl (i1,i2) = Word.toInt (Word.<< (Word.fromInt i1,Word.fromInt i2))
fun shr (i1,i2) = Word.toInt (Word.>> (Word.fromInt i1,Word.fromInt i2))

datatype z = datatype Color.t

(* The following two functions give the encoding of colours, both code in
   terms of bits 10-13 *)

fun colour_to_bits c : int =
  case c of
      Black   => 0x0400
    | Blue    => 0x0800
    | Cyan    => 0x0c00
    | Green   => 0x1000
    | Magenta => 0x1400
    | Red     => 0x1800
    | Unknown => 0x1c00
    | White   => 0x0000
    | Yellow  => 0x2000

fun colour_from_bits i =
    case andb(i,0x3c00) of
	0x0400 => Black
      | 0x0800 => Blue
      | 0x0c00 => Cyan
      | 0x1000 => Green
      | 0x1400 => Magenta
      | 0x1800 => Red
      | 0x1c00 => Unknown
      | 0x0000 => White
      | 0x2000 => Yellow
      | _ => raise (Fail "bad colour")
		   
fun toInternal { c, p = {b, em, i, s, tt, u, size, color} } : internal =
    let val b = if b then 0x1000000 else 0
	val em = if em then 0x800000 else 0
	val i = if i then 0x400000 else 0
	val s = if s then 0x200000 else 0
	val tt = if tt then 0x100000 else 0
	val u = shl(u,18)
	val sz = case size of NONE => 0x30000 | SOME size => shl(size,14)
	val color = colour_to_bits color
	val c = Char.toInt c
    in  orb (orb(b, orb (em, orb (i, s))),
	     orb (tt, orb (u, orb (sz, orb (color, c)))))
    end

fun toExternal (internal : internal): External.t =
  let val b = andb(internal,0x1000000) <> 0
      val em = andb(internal,0x800000) <> 0
      val i = andb(internal,0x400000) <> 0
      val s = andb(internal,0x200000) <> 0
      val tt = andb(internal,0x100000) <> 0
      val u = andb(shr(internal,18),3)
      val size = andb(internal,0x3c000)
      val size = if size=0x30000 then NONE else SOME (shr(size,14))
      val color = colour_from_bits internal
      val c = Char.fromInt (andb(internal,255))
  in  { c=c, p = { b=b, em=em, i=i, s=s, tt=tt, u=u, size=size, color=color } }
  end

fun equals (m: t, m': t): bool =
   Array.equals (m, m', op =)

fun explainNonEquals (m: t, m': t, out): unit =
   let
      val n = Array.length m
      val n' = Array.length m'
   in
      if n <> n'
	 then Out.output (out, concat ["different lengths ",
				       Int.toString n, " ",
				       Int.toString n', "\n"])
      else
	 let
	    fun loop i =
	       if i < n
		  then
		     let
			val c = Array.sub (m, i)
			val c' = Array.sub (m', i)
		     in
			if c = c'
			   then loop (i + 1)
			else
			   (Out.output (out, concat ["differ at character ",
						     Int.toString i, "\n"])
			    ; Layout.outputl (External.layout (toExternal c), out)
			    ; Layout.outputl (External.layout (toExternal c'), out)
			    ; Layout.outputl (External.layout (toExternal (Array.sub (m,i-1))), out)
			    ; Layout.outputl (External.layout (toExternal (Array.sub (m',i-1))), out)
			      )
		     end
	       else ()
	 in
	    loop 0
	 end
   end

fun layout (a: t): Layout.t =
   Array.layout (External.layout o toExternal) a

val layoutDecchar = External.layout

fun testAll () =
    let fun do_test (c,b,em,i,s,tt,u,size,color) =
	    let
	       val p = { b=b, em=em, i=i, s=s, tt=tt, u=u,
			size=size, color=color }
	       val d = { c = Char.fromInt c,
			 p =  p}
		val i' = toInternal d
		val d' = toExternal i'
	    in
	       if d<>d'
		  then
		     raise (Fail (concat ["tests failed for ",
					  Proprec.toString p]))
	       else ()
	    end
	fun do_char (b,em,i,s,tt,u,size,color) =
	    let fun aux c =
		    if c<127 then
			(do_test (c,b,em,i,s,tt,u,size,color);
			 aux (c+1))
		    else ()
	    in  (do_test (9,b,em,i,s,tt,u,size,color);
		 do_test (10,b,em,i,s,tt,u,size,color);
		 do_test (13,b,em,i,s,tt,u,size,color);
		 aux 32)
	    end
	fun do_b (em,i,s,tt,u,size,color) =
	    (do_char (true,em,i,s,tt,u,size,color);
	     do_char (false,em,i,s,tt,u,size,color))
	fun do_em (i,s,tt,u,size,color) =
	    (do_b (true,i,s,tt,u,size,color);
	     do_b (false,i,s,tt,u,size,color))
	fun do_i (s,tt,u,size,color) =
	    (do_em (true,s,tt,u,size,color);
	     do_em (false,s,tt,u,size,color))
	fun do_s (tt,u,size,color) =
	    (do_i (true,tt,u,size,color);
	     do_i (false,tt,u,size,color))
	fun do_tt (u,size,color) =
	    (do_s (true,u,size,color);
	     do_s (false,u,size,color))
	fun do_u (size,color) =
	    (do_tt (0,size,color);
	     do_tt (1,size,color);
	     do_tt (2,size,color);
	     do_tt (3,size,color))
	fun do_size color =
	    (do_u (NONE,color);
	     do_u (SOME 0,color);
	     do_u (SOME 1,color);
	     do_u (SOME 2,color);
	     do_u (SOME 3,color);
	     do_u (SOME 4,color);
	     do_u (SOME 5,color);
	     do_u (SOME 6,color);
	     do_u (SOME 7,color);
	     do_u (SOME 8,color);
	     do_u (SOME 9,color))
    in
       List.foreach
       ([Black, Blue, Cyan, Green, Magenta, Red, Unknown, White, Yellow],
	do_size)
    end

end
