functor Parse (S: PARSE_STRUCTS): PARSE = 
struct

open S

structure Color = Document.Color

structure Token =
   struct
      datatype t =
	 StartTag of Document.tag
       | EndTag of Document.tag
       | Text of string
       | EOF

      val toString =
	 fn StartTag t => Document.startTagToString t
	  | EndTag t => Document.endTagToString t
	  | Text s => s
	  | EOF => ""

      val layout = Layout.str o toString

      val isEof = fn EOF => true | _ => false
   end

datatype z = datatype Token.t
   
structure Array = Word8Array

fun tokenize (f: File.t): unit -> Token.t =
   let
      val buf = Array.array (4096, 0w0)
      val fd = let open Posix.FileSys
	       in openf (f, O_RDONLY, O.flags [])
	       end
      fun fillBuf () =
	 Int32.fromInt (Posix.IO.readArr (fd, {buf = buf,
					       i = 0,
					       sz = NONE}))
      val firstRef = ref 0
      val lastRef = ref 0 (* one past last char *)
   in
      Trace.trace ("nextToken", Unit.layout, Token.layout)
      (fn () =>
      let
	 fun safeGetChar (first, last): char * int * int =
	    (Char.fromWord8 (Array.sub (buf, first)), first + 1, last)
	 fun getChar (first, last) =
	    if first < last
	       then safeGetChar (first, last)
	    else
	       let
		  val n = fillBuf ()
	       in
		  if n = 0
		     then Error.bug "unexpected document end"
		  else safeGetChar (0, n)
	       end
	 fun finishTag (token, first, last) =
	    let
	       val (c, first, last) = getChar (first, last)
	    in
	       case c of
		  #">" => (firstRef := first
			   ; lastRef := last
			   ; token)
		| _ => Error.bug "expected to finish tag"
	    end
	 fun finishTag2 (c, token, first, last) =
	    let
	       val (c', first, last) = getChar (first, last)
	       val _ =
		  if c = c'
		     then ()
		  else Error.bug "bad tag"
	    in
	       finishTag (token, first, last)
	    end
	 fun tag (first, last) =
	    let
	       val (c, first, last) = getChar (first, last)
	       fun f t = finishTag (StartTag t, first, last)
	       fun f2 (c, t) = finishTag2 (c, StartTag t, first, last)
	       datatype z = datatype Document.tag
	       datatype z = datatype Color.t
	    in		  
	       case c of
		  #"B" => f B
		| #"E" => f2 (#"M", EM)
		| #"I" => f I
		| #"P" => f2 (#"L", PL)
		| #"S" => f S
		| #"T" => f2 (#"T", TT)
		| #"U" => f U
		| #"0" => f (Size 0)
		| #"1" => f (Size 1)
		| #"2" => f (Size 2)
		| #"3" => f (Size 3)
		| #"4" => f (Size 4)
		| #"5" => f (Size 5)
		| #"6" => f (Size 6)
		| #"7" => f (Size 7)
		| #"8" => f (Size 8)
		| #"9" => f (Size 9)
		| #"r" => f (Color Red)
		| #"g" => f (Color Green)
		| #"b" => f (Color Blue)
		| #"c" => f (Color Cyan)
		| #"m" => f (Color Magenta)
		| #"y" => f (Color Yellow)
		| #"k" => f (Color Black)
		| #"w" => f (Color White)
		| #"/" =>
		     let
			val (c, first, last) = getChar (first, last)
			fun f t = finishTag (EndTag t, first, last)
			fun f2 (c, t) = finishTag2 (c, EndTag t, first, last)
		     in
			case c of 
			   #"B" => f B
			 | #"E" => f2 (#"M", EM)
			 | #"I" => f I
			 | #"P" => f2 (#"L", PL)
			 | #"S" => f S
			 | #"T" => f2 (#"T", TT)
			 | #"U" => f U
			 | #"0" => f (Size 0)
			 | #"1" => f (Size 1)
			 | #"2" => f (Size 2)
			 | #"3" => f (Size 3)
			 | #"4" => f (Size 4)
			 | #"5" => f (Size 5)
			 | #"6" => f (Size 6)
			 | #"7" => f (Size 7)
			 | #"8" => f (Size 8)
			 | #"9" => f (Size 9)
			 | #"r" => f (Color Red)
			 | #"g" => f (Color Green)
			 | #"b" => f (Color Blue)
			 | #"c" => f (Color Cyan)
			 | #"m" => f (Color Magenta)
			 | #"y" => f (Color Yellow)
			 | #"k" => f (Color Black)
			 | #"w" => f (Color White)
			 | _ => Error.bug "strange end tag"
		     end
		 | _ => Error.bug "strange start tag"
	    end
	 fun loop (first, last, ac) =
	    let
	       fun loop' (i: int, ac) =
		  let
		     fun extract () =
			Byte.unpackString (buf, Int32.toInt first,
					   SOME (Int32.toInt (i - first)))
		  in
		     if i < last
			then if Char.toWord8 #"<" <> Array.sub (buf, i)
				then loop' (i + 1, ac)
			     else
				let
				   fun text s =
				     let
					val _ = firstRef := i
					val _ = lastRef := last
				     in
					Text s
				     end
				in
				   case ac of
				      [] =>
					 if i = first
					    then tag (i + 1, last)
					 else text (extract ())
				    | _ =>
					 text (concat (List.appendRev
						       (ac, [extract ()])))
				end
		     else
			let
			   val ac =
			      if i = first
				 then ac
			      else extract () :: ac
			   val n = fillBuf ()
			in
			   if n = 0
			      then (case ac of
				       [] => (Posix.IO.close fd
					      ; EOF)
				     | _ =>
					  (firstRef := 0
					   ; lastRef := 0
					   ; Text (concat (rev ac))))
			   else loop (0, n, ac)
			end
		  end
	    in
	       loop' (first, ac)
	    end
      in
	 loop (!firstRef, !lastRef, [])
      end
   )
   end

structure Token =
   struct
      open Token

      val tokenize = tokenize
   end

fun strip (f: File.t) : unit
  =
  let
     val nextToken = tokenize f

     fun strip' ()
       = case nextToken ()
	   of StartTag t
	    => let
		 val st = Document.startTagToString t
	       in
		 Out.output(Out.standard, st);
		 strip' ()
	       end
            | EndTag t
	    => let
		 val et = Document.endTagToString t
	       in
		 Out.output(Out.standard, et);
		 strip' ()
	       end
	    | Text s => strip' ()
	    | EOF => ()
  in
     strip' ()
  end

type ctxt = (Document.tag option * Document.item list) list
fun parse (f: File.t) : Document.t
  =
  let
     val nextToken = tokenize f

      fun makeDoc ds = Document.Doc (List.rev ds)
	
      fun pushText (s, ctxt : ctxt) : ctxt
	= (case ctxt
	     of (t,dis)::ctxt' 
	      => (t, (Document.Text s)::dis)::ctxt'
	      | _ => Error.bug "Parser:: pushText: empty context")

      fun pushTagged ((t',dis'), ctxt : ctxt) : ctxt
	= (case ctxt
	     of (t,dis)::ctxt' 
	      => (t, (Document.Tagged (t', makeDoc dis'))::dis)::ctxt'
	      | _ => Error.bug "Parser:: pushTagged: empty context")

      fun parse' (ctxt : ctxt) : Document.t
	= case nextToken ()
	    of StartTag t 
	     => parse' ((SOME t,[])::ctxt)
	     | EndTag t
	     => (case ctxt
		   of (SOME t',dis')::ctxt'
		    => if t = t'
			 then parse' (pushTagged ((t',dis'), ctxt'))
			 else Error.bug "Parser:: mismatch at EndTag"
		    | _ => Error.bug "Parser:: empty context at EndTag")
	     | Text s 
	     => parse' (pushText (s, ctxt))
	     | EOF
	     => (case ctxt
		   of (NONE,dis)::nil => makeDoc dis
		    | _ => Error.bug "Parser:: non-empty context at EOF")
    in
      parse' [(NONE,[])]
    end

end
